////////////////////////////////////////////////////////////////////////////////
// File: tridiagonalize_symmetric_matrix.c                                    //
// Routine(s):                                                                //
//    Tridiagonalize_Symmetric_Matrix                                         //
//                                                                            //
// Required Externally Defined Routines:                                      //
//    Copy_Matrix                                                             //
//    Inner_Product                                                           //
////////////////////////////////////////////////////////////////////////////////
#include <math.h>                 // required for sqrt()
#include <float.h>                // required for DBL_MIN and DBL_EPSILON

double Inner_Product(double u[], double v[], int n)
{
   int i;
   double sum;
   sum = 0.0;
   for(i=0; i<n; i++)  sum += u[i] * v[i];
   return(sum);
}

////////////////////////////////////////////////////////////////////////////////
//  int Tridiagonalize_Symmetric_Matrix(double *A, int storage_class,         //
//               double diagonal[], double off_diagonal[], double *U, int n)  //
//                                                                            //
//  Description:                                                              //
//     This program transforms the symmetric square matrix A to a similar     //
//     tridiagonal matrix by a multiplying A on the right and left by a       //
//     sequence of Householder transformations.                               //
//     Def:  Two matrices A and B are said to be similar if there exists a    //
//           nonsingular matrix S such that A S = S B.                        //
//     Def:  A Householder transformation is an orthogonal transformation of  //
//           the form Q = I - 2 uu'/u'u, where u is a n x 1 column matrix and //
//           ' denotes the transpose.                                         //
//     Thm:  If Q is a Householder transformation then Q' = Q  and  Q' Q = I, //
//           i.e. Q is a symmetric involution.                                //
//     First the input matrix A is copied / expanded to form the matrix U.    //
//     The algorithm proceeds by successivly selecting rows i = n - 1, ..., 1 //
//     and then calculating the Householder transformation Q which annihilates//
//     the components to the left of the subdiagonal for that row and leaves  //
//     the previously selected rows invariant.  The algorithm then updates    //
//     the matrix U, in place, by premultiplication by Q followed by          //
//     postmultiplication by Q.                                               //
//     If the i-th row of U is (a[0],...,a[n-1]), then  choose u' =           //
//     (u[0],...,u[n-1]) by u[0] = 0, ... , u[j] = 0, u[j+2] = a[j+2],...,    //
//     u[n-1] = a[n-1].  The remaining component u[j+1] = a[j+1] - s, where   //
//     s^2 = a[j+1]^2 + ... + a[n-1]^2, and the choice of sign for s,         //
//     sign(s) = -sign(a[j+1]) maximizes the number of significant bits for   //
//     u[j+1].                                                                //
//                                                                            //
//     Note: In general accuracy is improved rows and columns of the matrix   //
//     are ordered so that the elements with the smallest magnitude are       //
//     located at the top left-most corner of the matrix and the elements with//
//     the largest magnitude are located at the bottom right-most corner.     //
//                                                                            //
//  Arguments:                                                                //
//     double *A                                                              //
//            Pointer to the first element of the matrix A[n][n].             //
//     int    storage_class                                                   //
//            The storage class of the matrix A, 0 if full symmetric matrix   //
//            -1 if stored in lower triangular form, +1 if stored in upper    //
//            triangular form.                                                //
//     double diagonal[]                                                      //
//            The diagonal of the tridiagonal symmetric matrix.  The array    //
//            diagonal[] should be dimensioned n in the calling routine.      //
//     double off_diagonal[]                                                  //
//            The subdiagonal and superdiagonal of the tridiagonal symmetric  //
//            matrix.  The subdiagonal and superdiagonal are (off_diagonal[1],//
//            ,,, off_diagonal[n-1]), off_diagonal[0] is set to 0.0.  The     //
//            array off_diagonal[] should be dimensioned n in the calling     //
//            routine.  If QL_Tridiagonal_Symmetric_Matrix is called using    //
//            using diagonal[], off_diagonal[], and U, then off_diagonal      //
//            should be dimensioned n + 1 in the calling routine.             //
//     double *U                                                              //
//            The n x n orthogonal matrix such that AU = UT, where T is the   //
//            tridiagonal matrix with diagonal 'diagonal[]' and off-diagonal  //
//            'off_diagonal[]'.                                               //
//     int    n                                                               //
//            The number of rows or columns of the symmetric matrix A and of  //
//            the matrix U.                                                   //
//                                                                            //
//  Return Values:                                                            //
//     0  Success                                                             //
//    -1  Failure - Illegal storage class.                                    //
//                                                                            //
//  Example:                                                                  //
//     #define N                                                              //
//     double A[N][N], B[N * (N + 1) / 2], U[N][N], diagonal[N];              //
//     double off_diagonal[N];                                                //
//                                                                            //
//     (your code to create the matrices A and B)                             //
//     Tridiagonalize_Symmetric_Matrix((double*)A, 0, diagonal, off_diagonal, //
//                                                          (double*)U, n);   //
//     (if B is the matrix stored in lower triangular form )                  //
//     Tridiagonalize_Symmetric_Matrix((double*)B, -1, diagonal, off_diagonal,//
//                                                          (double*)U, n);   //
//     (if B is the matrix stored in upper triangular form )                  //
//     Tridiagonalize_Symmetric_Matrix((double*)B, 1, diagonal, off_diagonal, //
//                                                          (double*)U, n);   //
////////////////////////////////////////////////////////////////////////////////
//                                                                            //
int Tridiagonalize_Symmetric_Matrix(MATRIX *mat_A,
                   double diagonal[], double off_diagonal[], MATRIX *mat_U, int n)
{
   int i, j, k, col;
   double zero_tolerance = DBL_MIN / DBL_EPSILON; 
   double *U, *A, *pU, *ppU, *pUrow, *pUcol;
   double ss;                             // signed sqrt of sum of squares

   U = mat_U->data[0];
   A = mat_A->data[0];
       // n x n matrices for which n <= 2 are already in Tridiagonal form

   if (n <= 2) { 
      for (i = 0, k = 0; i < n; i++) { 
         for (j = 0; j < i; j++, k++) *(U + k) = 0.0;
         *(U + k++) = 1.0;
         for (j = i + 1; j < n; j++, k++) *(U + k) = 0.0;
      } 
      return 0;
   }

   duplicate(mat_A, mat_U);

           // For each row starting with the last row find the 
           // Householder transformation to zero all entries to
           // the left of the subdiagonal.

   for (i = n - 1, pUrow = U + n * (n - 1); i > 0; i--, pUrow -= n) {

          // Calculate the sum of squares of the elements to the left
          // of the subdiagonal.
      
      pU = pUrow + i - 1;
      ss = Inner_Product(pUrow, pUrow, i - 1);

          // If the sum of squares is too small set, don't transform.

      if (ss < zero_tolerance) {
         off_diagonal[i] = *pU;
         diagonal[i] = 0.0;
         continue;
      }

         // Calculate the signed square root of the sum of squares to the
         // left of the diagonal.

      ss += *pU * *pU;
      off_diagonal[i] = (*pU >= 0.0) ? -sqrt(ss) : sqrt(ss);
      diagonal[i] = ss - *pU * off_diagonal[i];
      *pU -= off_diagonal[i];
      for (j = 0, pU = U, ss = 0.0; j < i; j++, pU += n) {
         *(pU + i) = *(pUrow +j) / diagonal[i];
         off_diagonal[j] = Inner_Product(pU, pUrow, j + 1);
         for (k = j+1, ppU = pU + n; k < i; k++, ppU += n)
            off_diagonal[j] += *(ppU + j) * *(pUrow + k);
         ss += off_diagonal[j] * *(pU + i);
         off_diagonal[j] /= diagonal[i];
      } 
      ss /= (diagonal[i] + diagonal[i]);
      for (j = 0, pU = U; j < i; j++, pU += n) {
        off_diagonal[j] -= ss * *(pUrow + j);
        for (k = 0; k <= j; k++)
           *(pU + k) -= ( *(pUrow + j) * off_diagonal[k]
                          + off_diagonal[j] * *(pUrow + k) );
      }
   }
   diagonal[0] = 0.0;
   off_diagonal[0] = 0.0; 
   for (i = 0, pUrow = U; i < n; i++, pUrow += n) {
      if ( diagonal[i] != 0.0 ) 
         for (j = 0; j < i; j++) {
            for (k = 0, pU = U, ss = 0.0; k < i; k++, pU += n)
               ss += *(pUrow + k) * *(pU + j);
            for (k = 0, pU = U; k < i; k++, pU += n)
               *(pU + j) -= ss * *(pU + i);
         }
      diagonal[i] = *(pUrow + i);
      *(pUrow + i) = 1.0;
      for (j = 0, pU = U; j < i; j++, pU += n) {
         *(pUrow + j) = 0.0;
         *(pU + i) = 0.0;
      }
   }    
   return 0;
}





////////////////////////////////////////////////////////////////////////////////
// File: ql_tridiagonal_symmetric_matrix.c                                    //
// Routine(s):                                                                //
//    QL_Tridiagonal_Symmetric_Matrix                                         //
//                                                                            //
// Required Externally Defined Routines:                                      //
//    Copy_Matrix                                                             //
//    Inner_Product                                                           //
////////////////////////////////////////////////////////////////////////////////


//                        Internally Defined Routines 
static double Calculate_Shift(double d2, double d1, double off); 
static double Transform_Matrix(double* U, double s, double c, int col, int n);
////////////////////////////////////////////////////////////////////////////////
//  int QL_Tridiagonal_Symmetric_Matrix(double diagonal[],                    //
//          double off_diagonal[], double *U, int n, int max_iteration_count) //
//                                                                            //
//  Description:                                                              //
//     This function implements the QL algorithm with implicit shifts of the  //
//     of the origin for a symmetric tridiagonal matrix.  If the original     //
//     symmetric matrix was tridiagonalized using Householder transformations //
//     then the tridiagonal matrix should be organized so that the largest    //
//     elements are at the bottom right-hand of the matrix.  Also, the process//
//     of tridiagonalizing the matrix will introduce round-off errors, so that//
//     the estimates of the eigenvalues and eigenvectors may be improved by   //
//     calling the inverse_power_method.                                      //
//                                                                            //
//  Arguments:                                                                //
//     double diagonal[]                                                      //
//            On input, the diagonal of the tridiagonal symmetric matrix.     //
//            On output, the eigenvalues.                                     //
//     double off_diagonal[]                                                  //
//            The subdiagonal and superdiagonal of the tridiagonal symmetric  //
//            matrix.  The subdiagonal and superdiagonal are (off_diagonal[1],//
//            ,,, off_diagonal[n-1]).                                         //
//     double *U                                                              //
//            On input, U[n][n] is the identity matrix if the tridiagonal     //
//            matrix is the primary data or is the transformation matrix      //
//            of Householder transformations which is the output of           //
//            tridiagonalize_symmetric_matrix.  On output, U[n][n] is the     //
//            matrix of eigenvectors, the ith column being the eigenvector    //
//            corresponding to the eigenvalue d[i].  These are the eigen-     //
//            vectors of the full matrix, if on input U is not the identity.  //
//     int    n                                                               //
//            The number of rows or columns of the symmetric matrix A and of  //
//            the matrix U.                                                   //
//     int    max_iteration_count                                             //
//            The maximum number of iterations to try to annihilate an off-   //
//            diagonal elements.                                              //
//                                                                            //
//  Return Values:                                                            //
//     0  Success                                                             //
//    -1  Failure - Failed to converge within max_iteration_count iterations. //
//                                                                            //
//  Example:                                                                  //
//     #define N                                                              //
//     double U[N][N], diagonal[N], double off_diagonal[N];                   //
//     int err;                                                               //
//     int max_iteration_count = 30;                                          //
//                                                                            //
//     (your code to create the vectors diagonal and off_diagonal)            //
//     if (QL_Tridiagonal_Symmetric_Matrix(diagonal, off_diagonal, (double*)U,//
//         N, max_iteration_count) ) printf("Failed to converge\n");          //
//     else printf("Success\n");                                              //
////////////////////////////////////////////////////////////////////////////////
//                                                                            //
int QL_Tridiagonal_Symmetric_Matrix( double diagonal[], double off_diagonal[],
                                   MATRIX *mat_U, int n, int max_iteration_count)
{
   int i, j, k;
   int iteration;
   double epsilon; 
   double *p_off = off_diagonal + 1;
   double s,c,g,h,q;
   double shift;
   double dum;
   double *U;

   U = mat_U->data[0];
   
   for(i=0; i<n; i++)
   {
      mat_U->data[i][i] = 1.0;
      for(j=i+1; j<n; j++)
      {
         mat_U->data[i][j] = 0.0;
         mat_U->data[j][i] = 0.0;
      }
   }
      
   p_off[n-1] = 0.0;
   for (i = 0; i < n; i++) 
   {
      printf("i=%d\n", i);
      for (iteration = 0; iteration < max_iteration_count; iteration++) 
      { 
         printf("iteration=%d\n", iteration);
         for (k = i; k < (n - 1); k++) 
         {
            epsilon = DBL_EPSILON * ( fabs(diagonal[k]) + fabs(diagonal[k+1]) );
            if ( fabs(p_off[k]) <= epsilon ) break;
         }
         if ( k == i ) break;
         shift = Calculate_Shift(diagonal[i+1], diagonal[i], p_off[i]);
         q = diagonal[k] - shift;
         c = 1.0;
         s = 1.0;
         shift = 0.0;
         for (j = k - 1; j >= i; j--) 
         {
            h = c * p_off[j];
            g = s * p_off[j];
            if ( fabs( g ) >= fabs( q ) ) 
            {
               c = q / g;
               dum = sqrt( c * c + 1.0 );
               p_off[j+1] = g * dum;
               s = 1.0 / dum;
               c *= s;
            }
            else 
            {
               s = g / q;
               dum = sqrt( s * s + 1.0 );
               p_off[j+1] = q * dum;
               c = 1.0 / dum;
               s *= c;
            }
            q = diagonal[j + 1] - shift;
            dum = s * (diagonal[j] - q) + 2.0 * c * h;
            shift = s * dum;
            diagonal[j+1] = q + shift;
            q = c * dum - h;
            Transform_Matrix(U, s, c, j, n);
         }
         diagonal[i] -= shift;
         p_off[i] = q;
         p_off[k] = 0.0;
      }
   }
   printf("iteration=%d\n", iteration);
   if ( iteration >= max_iteration_count ) return -1;    
   return 0;
}

static double Calculate_Shift(double d2, double d1, double off) 
{
   double h;

   h = ( d2 - d1 ) / ( off + off );
   off = sqrt( h * h + 1.0 );
   d2 = ( h < 0.0 ) ? h - off : h + off;
   return d1 - off / d2;
}

static double Transform_Matrix(double* U, double s, double c, int col, int n)
{
   double x;
   int i;
   int col1 = col + 1;

   for (i = 0; i < n; i++, U += n) {
      x = *(U + col1);
      *(U + col1) = s * *(U + col) + c * x;
      *(U + col) *= c;
      *(U + col) -= s * x;
   }
}

#define ROTATE(a,i,j,k,l) g=a[i][j];h=a[k][l];a[i][j]=g-s*(h+g*tau);a[k][l]=h+s*(g-h*tau);

int eigen(MATRIX *a, int n, double *d, MATRIX *v, int *nrot)
{
   int j, iq, ip, i;
   double tresh, theta, tau, t, sm, s, h, g, c, *b, *z;
   b = (double *) malloc((size_t) n * sizeof(double));
   z = (double *) malloc((size_t) n * sizeof(double));
   for(ip=0; ip<n; ip++)
   {
      for(iq=0; iq<n; iq++)
         v->data[ip][iq]=0.0;
      v->data[ip][ip]=1.0;
   }
   for(ip=0; ip<n; ip++)
   {
      b[ip] = d[ip] = a->data[ip][ip];
      z[ip] = 0.0;
   }
   *nrot = 0;
   for(i=1; i<=100; i++)
   {
      sm=0.0;
      for(ip=0; ip<n-1; ip++)
      {
         for(iq=ip+1; iq<n; iq++)
           sm += fabs(a->data[ip][iq]);
      }
      if(sm == 0.0)
      {
         free(z);
         free(b);
         return(0);
      }
      if(i < 4)  tresh = 0.2 * sm / (n * n);
      else  tresh = 0.0;
      for(ip=0; ip<n-1; ip++)
      {
         for(iq=ip+1; iq<n; iq++)
         {
            g = 100.0 * fabs(a->data[ip][iq]);
            if(i > 4 && (fabs(d[ip]) + g) == fabs(d[ip])
                     && (fabs(d[iq]) + g) == fabs(d[iq]))
               a->data[ip][iq] = 0.0;
            else if(fabs(a->data[ip][iq]) > tresh)
            {
               h = d[iq] - d[ip];
               if((double) (fabs(h) + g) == (double) fabs(h))
                  t = a->data[ip][iq] / h;
               else
               {
                  theta = 0.5 * h / a->data[ip][iq];
                  t = 1.0 / (fabs(theta) + sqrt(1.0 + theta * theta));
                  if(theta < 0.0)  t = -t;
               }
               c = 1.0 / sqrt(1 + t * t);
               s = t * c;
               tau = s / (1.0 + c);
               h = t * a->data[ip][iq];
               z[ip] -= h;
               z[iq] += h;
               d[ip] -= h;
               d[iq] += h;
               a->data[ip][iq] = 0.0;
               for(j=0; j<=ip-1; j++)  
               {
                  /*printf("ip=%d  iq=%d  j=%d\n", ip, iq, j);*/
                  ROTATE(a->data, j, ip, j, iq)
               }
               for(j=ip+1; j<=iq-1; j++)  
               {
                  /*printf("ip=%d  iq=%d  j=%d\n", ip, iq, j);*/
                  ROTATE(a->data, ip, j, j, iq)
               }
               for(j=iq+1; j<n; j++)  
               {
                  /*printf("ip=%d  iq=%d  j=%d\n", ip, iq, j);*/
                  ROTATE(a->data, ip, j, iq, j)
               }
               for(j=0; j<n; j++)  
               {
                  /*printf("ip=%d  iq=%d  j=%d\n", ip, iq, j);*/
                  ROTATE(v->data, j, ip, j, iq)
               }
               ++(*nrot);
            }
         }
      }
      for(ip=0; ip<n; ip++)
      {
         b[ip] += z[ip];
         d[ip] = b[ip];
         z[ip] = 0.0;
      }
   }
   free(z);
   free(b);
   return(1);
}
